/*
 *  Copyright (c) by Uros Bizjak <uros@kss-loka.si>
 *
 *  lowlevel code for SAM9407 chip - microcode hwdep interface
 *
 *  BUGS:
 *    --
 *
 *  TODO:
 *    --
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#define __NO_VERSION__
#include <sound/driver.h>
#include <asm/io.h>
#include <linux/delay.h>
#include <linux/wait.h>
#include <sound/core.h>
#include <sam9407.h>

#define SAM_UCODE_OFFS		0x200


#include "sam_bootcode.h"

static int snd_sam9407_info(sam9407_t * sam, sam9407_info_t *info)
{
	// TODO
	snd_assert(sam != NULL, {snd_printk("sam = NULL !");return -1;});
	snd_assert(info != NULL, {snd_printk("info = NULL !");return -1;});
	info->empty = 0;
	return 0;
}

static int snd_sam9407_ucode_load(sam9407_t *sam, sam9407_ucode_t *ucode)
{
	u8 bytes[7];

	int err;

	//u16 io_port = 0xB400;


	/* FIXME: if PCM devices are present, close them  ??*/

	//snd_printk(" Init host enviroment\n");


	//ucode->len -= SAM_UCODE_OFFS;
	if (sam) {
		if (sam->init(sam)) {
			snd_printk("Problem initialisising host enviroment");
		}
	}
	
  

	if(sam->ucode_loaded) {
#if 0
		/* remove module to reload microcode */
		return -EBUSY;
#else
#if 0
		/* reset loaded microcode - doesn't work on EWS64 */
		err = snd_sam9407_command(sam, NULL, SAM_CMD_EN_CONTROL,
					  NULL, 0, NULL, 0, -1);
		bytes[0] = SAM_DATA_HOT_RES;	/* Hot reset */
		err = snd_sam9407_command(sam, sam->synth_io_queue, SAM_CMD_HOT_RES,
					  bytes, 1, NULL, 0, 0x00);
		if (err < 0) {
			snd_printd("sam9407: hot reset failed\n");
			return err;
		}
#else
	/* reset the device */
	sam->writeCtrl(sam,0x70);
	sam->writeData8(sam,0x11);
	udelay(1000);
	sam->readData8(sam);
	/* should be done now */
#endif
#endif
	}
	//snd_printk(" Send bootcode\n");

	/* send boot microcode */
	sam->writeData16Burst(sam, snd_sam9407_boot_ucode, SAM_ARRAYSIZE(snd_sam9407_boot_ucode));
	//outsw(SAM9407P(sam, DATA16), snd_sam9407_boot_ucode, SAM_ARRAYSIZE(snd_sam9407_boot_ucode));

	//snd_printk(" bootcode sent\n");
 

	/* wait for boot code to start */
	udelay(1000);

	/* here the isis does some strange things */


	//snd_printk(" strange ctrl 0x04 \n");

	sam->writeCtrl(sam,0x04);
	sam->writeCtrl(sam,0x00);
	udelay(1000);

	while(!(sam->readCtrl(sam) & 0x80)) {
		snd_printk(" cmd 0x04 respose: %x\n",sam->readData8(sam));
	}


	//snd_printk(" strange control 0x05\n");


	sam->writeCtrl(sam,0x05);
	sam->writeCtrl(sam,0x00);
	sam->writeCtrl(sam,0x00);
	sam->writeCtrl(sam,0x00);

	udelay(1000);
	//snd_printk(" start microcode transfer\n");
  


	/* send "transfer microcode" command sequence to boot microcode */
	bytes[0] = SAM_BOOT_CMD_UCODE_TRANSFER;	/* transfer microcode "command" */
	bytes[1] = (u8) SAM_UCODE_OFFS;	/* address offset */
	bytes[2] = (u8) (SAM_UCODE_OFFS >> 8);
	bytes[3] = 0x00;			/* address page */
	bytes[4] = 0x00;
	bytes[5] = (u8) ucode->len;		/* xfer length */
	bytes[6] = (u8) (ucode->len >> 8);
	err = snd_sam9407_init_command(sam, NULL, bytes, 7, NULL, 0, -1);
	if (err)
		goto __error;
 

	udelay(1000);
  
	while(!(sam->readCtrl(sam) & 0x80)) {
		snd_printk(" cmd respose: %x\n",sam->readData8(sam));
	}
 

	/* the windows driver reads response first */
	//snd_printk("  command sent. starting write\n");

	/* short delay is needed here */
	udelay(300);
	//udelay(30);
 
	/* send microcode */
	sam->writeData16Burst(sam, ucode->prog, ucode->len);
//	sam->writeData16Burst(sam, ucode->prog + SAM_UCODE_OFFS, ucode->len);
	//outsw(SAM9407P(sam, DATA16), ucode->prog + SAM_UCODE_OFFS, ucode->len);
 

	//snd_printk(" write done. starting microcode \n");

	/* send "start microcode" command sequence to boot code */
	bytes[0] = SAM_BOOT_CMD_UCODE_START;	/* start microcode "command" */
	bytes[1] = (u8) SAM_UCODE_OFFS;	/* start address offset */
	bytes[2] = (u8) (SAM_UCODE_OFFS >> 8);
	// the ack is from the MIDI device and = 0
	// no bytes are read besides the ack.
	err = snd_sam9407_init_command(sam, NULL,
				       bytes, 3, NULL, 0, -1);
//	err = snd_sam9407_init_command(sam, sam->system_io_queue,
//				       bytes, 3, buffer, 2, SAM_ACK_CKSUM);
 
	sam->enableInterrupts(sam);
 	mdelay(100);


 __error:
	if (err < 0) {
		snd_printd("sam9407: start microcode command failed\n");
		return err;
	}
/*
	csum = (buffer[1] << 8) | buffer[0];
	if (csum != ucode->csum) {
		snd_printk("sam9407: microcode checksum failure\n");
		return err;
	}
*/
	/* switch to UART mode */
	err = snd_sam9407_command(sam, sam->midi_io_queue, SAM_CMD_UART_MOD,
				  NULL, 0, NULL, 0, SAM_ACK_MOD);
	if (err < 0) {
		snd_printd("sam9407: can't switch to UART mode\n");
		return err;
	}
 


	/* test board by generating an interrupt */
	bytes[0] = 0x00;
	err = snd_sam9407_command(sam, sam->system_io_queue, SAM_CMD_GEN_INT,
				  bytes, 1, NULL, 0, SAM_ACK_IRQ);
	if (err < 0) {
		snd_printd("sam9407: interrupt not detected\n");
		return err;
	}

	/* initialize chip */
	err = snd_sam9407_init(sam);
	if (err < 0)
		return err;

	sam->ucode_loaded = 1;

	return 0;

}


static int snd_sam9407_hwdep_open(snd_hwdep_t * hw, struct file *file)
{
 	sam9407_t *sam;

	snd_printk(" HWDEP open\n");
	snd_assert(hw != NULL, {snd_printk("hw = NULL !");return -1;});
	snd_assert(file != NULL, {snd_printk("file = NULL !");return -1;});

	sam = snd_magic_cast(sam9407_t, hw->private_data, return -ENXIO);
	snd_assert(sam != NULL, {snd_printk("sam = NULL !");return -1;});

	/* opened in exclusive mode */
	down(&sam->access_mutex);
	if (sam->hwdep_used) {
		up(&sam->access_mutex);
		return -EAGAIN;
	}
	sam->hwdep_used++;
	up(&sam->access_mutex);

	return 0;
}

static int snd_sam9407_hwdep_release(snd_hwdep_t * hw, struct file *file)
{
 	sam9407_t *sam;
	snd_printk(" HWDEP release\n");
	sam = snd_magic_cast(sam9407_t, hw->private_data, return -ENXIO);

	down(&sam->access_mutex);
	sam->hwdep_used--;
	up(&sam->access_mutex);

	return 0;
}

static int snd_sam9407_hwdep_ioctl(snd_hwdep_t * hw, struct file *file, unsigned int cmd, unsigned long arg)
{
 	sam9407_t *sam;
	sam9407_info_t info;
	sam9407_ucode_t *ucode;
	int res;

	snd_printk(" HWDEP ioctl\n");
	snd_assert(hw != NULL, {snd_printk("hw = NULL !");return -1;});
	snd_assert(file != NULL, {snd_printk("file = NULL !");return -1;});
	snd_assert((void *)arg != NULL, {snd_printk("arg = NULL !");return -1;});

	sam = snd_magic_cast(sam9407_t, hw->private_data, return -ENXIO);

	snd_assert(sam != NULL, {snd_printk("sam = NULL !");return -1;});

	switch (cmd) {
	case SNDRV_SAM9407_IOCTL_INFO:
		if ((res = snd_sam9407_info(sam, &info)) < 0)
			return res;
		if (copy_to_user((void *)arg, &info, sizeof(info)))
			return -EFAULT;
		return 0;
	case SNDRV_SAM9407_IOCTL_FW_LOAD:
#if 0
		if (!capable(CAP_SYS_ADMIN))
			return -EPERM;
#endif
		ucode = (sam9407_ucode_t *)vmalloc(sizeof(*ucode));
		if (ucode == NULL)
			return -ENOMEM;
		if (copy_from_user(ucode, (void *)arg, sizeof(*ucode))) {
			vfree(ucode);
			return -EFAULT;
		}
		if ((ucode->len < SAM_UCODE_OFFS) || (ucode->len > SAM_MAX_UCODE_LEN)) {
			vfree(ucode);
			return -EINVAL;
		}
		
		snd_assert(sam != NULL, {snd_printk("sam = NULL !");return -1;});
    		snd_assert(ucode != NULL, {snd_printk("ucode = NULL !");return -1;});

		res = snd_sam9407_ucode_load(sam, ucode);
		vfree(ucode);
		return res;
	case SNDRV_SAM9407_IOCTL_SBK_LOAD:
		
		return 0;
	}
	return -ENOTTY;
}

int snd_sam9407_ucode_hwdep_new(sam9407_t *sam, int device)
{
	snd_hwdep_t *hw;
	int err;
 	snd_printk("new HWDEP\n");

	snd_assert(sam != NULL, {snd_printk("sam = NULL !");return -1;});
	snd_assert(sam->card != NULL, {snd_printk("sam->card = NULL !");return -1;});

	if ((err = snd_hwdep_new(sam->card, "SAM9407", device, &hw)) < 0)
		return err;
	strcpy(hw->name, "SAM9407");
	hw->iface = SNDRV_HWDEP_IFACE_SAM9407;
	hw->ops.open = snd_sam9407_hwdep_open;
	hw->ops.release = snd_sam9407_hwdep_release;
	hw->ops.ioctl = snd_sam9407_hwdep_ioctl;
	hw->private_data = sam;
	return 0;
}
